#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <inttypes.h>

#include "em_device.h"
#include "em_chip.h"
#include "em_emu.h"
#include "em_cmu.h"
#include "em_gpio.h"
#include "em_usart.h"
#include "bsp.h"

#include "visi_genie.h"
#include "rtcdriver.h"


/* Function prototypes */

void uartPutData(uint8_t * dataPtr, uint32_t dataLen);
uint32_t uartGetData(uint8_t * dataPtr, uint32_t dataLen);
void    uartPutChar(uint8_t charPtr);
uint8_t uartGetChar(void);

/* Declare a circular buffer structure to use for Rx and Tx queues */
#define BUFFERSIZE          512

volatile struct circularBuffer
{
  uint8_t  data[BUFFERSIZE];  /* data buffer */
  uint32_t rdI;               /* read index */
  uint32_t wrI;               /* write index */
  uint32_t pendingBytes;      /* count of how many bytes are not yet handled */
  bool     overflow;          /* buffer overflow indicator */
} rxBuf, txBuf = { {0}, 0, 0, 0, false };

volatile uint32_t msTicks; /* counts 1ms timeTicks */

void Delay(uint32_t dlyTicks);

/**************************************************************************//**
 * @brief SysTick_Handler
 * Interrupt Service Routine for system tick counter
 *****************************************************************************/
void SysTick_Handler(void)
{
  msTicks++;       /* increment counter necessary in Delay()*/
}

/**************************************************************************//**
 * @brief Delays number of msTick Systicks (typically 1 ms)
 * @param dlyTicks Number of ticks to delay
 *****************************************************************************/
void Delay(uint32_t dlyTicks)
{
  uint32_t curTicks;

  curTicks = msTicks;
  while ((msTicks - curTicks) < dlyTicks) ;
}

//Function prototypes
void    writec(char data); 
// void    RxIrqHandler(void);
void    ManageReceiveData(uint8_t data);
bool    _genieEnqueueEvent (uint8_t * data);
void    _genieFlushEventQueue(void);
/*
    Wait for the screen to acknowledge the command
    returns the status code of operation
*/
bool    WaitForIdle();

int8_t  WaitForAnswer();

int8_t  WaitForReadAnswer();


//////////////////////////////////////////////////////////////
// A structure to hold up to MAX_GENIE_EVENTS events receive
// from the display
//
struct genieEventQueueStruct _genieEventQueue;
uint8_t     rxframe_count;
uint8_t     rx_data[GENIE_FRAME_SIZE];
uint8_t     checksum;

//////////////////////////////////////////////////////////////
// Pointer to the user's event handler function
//
genieUserEventHandlerPtr _genieUserHandler;

USART_TypeDef *_screen;
RTCDRV_TimerID_t   _waitTimer, _receptionTimer;

/*
DigitalOut      *_reset;
,_receptionTimer;
*/
int             state; 
int             LastResponse; 
long            RxMaxTimeout;
long            RxStateTimeoutErrors;


void rtcCallback( RTCDRV_TimerID_t id, void *user)
{
	/* Reset Pin*/
	GPIO_PinOutSet(gpioPortB, 11);
}

void GenieInit(void) /* (USART_TypeDef *usart, DigitalOut *resetpin)*/
{
    _screen = uart;

    /* Reset Pin*/
	GPIO_PinModeSet(gpioPortB, 11, gpioModeInput, 1);

	/*_reset = resetpin;*/

	_genieUserHandler = NULL;
    RxStateTimeoutErrors = 0;
}

void Start(void)
{
	GenieInit();

	/* Reset Pin*/
	GPIO_PinOutClear(gpioPortB, 11);

	// Reserve a timer
	RTCDRV_AllocateTimer( &_waitTimer );

	RTCDRV_StartTimer( _waitTimer, rtcdrvTimerTypeOneshot, 0, rtcCallback, NULL );

	Delay (3000);

	/*_reset = 1;
    wait(3.0);  //4D data sheet says that the screen can take up to 3000 ms before
                //becoming responsive to serial commands.

    RTCDRV_StartTimer( _receptionTimer, rtcdrvTimerTypeOneshot, 0, rtcCallback, NULL );

    RTCDRV_StartTimer(_waitTimer);.start();
    RTCDRV_StartTimer(_receptionTimer).start();
    */

    Reset();
    _genieFlushEventQueue();
    UART1_RX_IRQHandler();
    /*_screen.attach(&RxIrqHandler,Serial::RxIrq);*/
}
////////////////////// genieGetEventData ////////////////////////
//
// Returns the LSB and MSB of the event's data combined into
// a single uint16
//
// The data is transmitted from the display in big-endian format 
// and stored the same so the user can't just access it as an int 
// directly from the structure. 
//
uint16_t genieGetEventData (union genieFrame * e) {
        return  (e->reportObject.data_msb << 8) + e->reportObject.data_lsb;
}

/*
void RxIrqHandler(void)
{
    uint8_t c;
    //Loop to read all byte present in UART FIFO
    do
    {
        c = uartGetChar();

            check if we are in reception mode and if we timeout

        if(milliseconds >= RxMaxTimeout && state == CommInProgress)
        {

                if we get here, it means something bad has happened and 
                we need to reset communication status

            Reset();
        }

        ManageReceiveData(c);
    }
    while(uartGetChar( ));
}
*/

void ManageReceiveData(uint8_t data)
{
    switch(state)
    {
        case CommIdle:
        
            if(data == GENIE_ACK || data == GENIE_NAK)
            {
                LastResponse = data;
            }   
            else if(data == GENIE_REPORT_OBJ || data == GENIE_REPORT_EVENT)
            {
                //re-initialise the reception timer
                /*
            	_receptionTimer.reset();
                RxMaxTimeout = _receptionTimer.read_ms() + RESYNC_PERIOD;
                */
                checksum = data;
                rx_data[rxframe_count++] = data;
                state = CommInProgress;              
            }                     
            break;
        
        case CommInProgress: 
            checksum = checksum ^ data;
            rx_data[rxframe_count++] = data;
            
            if(rxframe_count >= GENIE_FRAME_SIZE)
            {
                if (checksum == 0) {
                        _genieEnqueueEvent(rx_data); 
                        if(_genieUserHandler != NULL)
                        {
                            (_genieUserHandler)();
                        }
                    } 
                state = CommIdle;
                rxframe_count = 0;
                LastResponse = NO_RESPONSE;
            }        
            break;
        default:
            state = CommIdle;
            rxframe_count = 0;
            LastResponse = NO_RESPONSE;
            break;
    }
}

///////////////////////// genieWriteObject //////////////////////
//
// Write data to an object on the display
//
uint8_t genieWriteObject (uint16_t object, uint16_t index, uint16_t data)
{
    uint16_t msb, lsb ;
    uint8_t checksum ;
    
    //wait for interface to be ready before sending stuff
    if(WaitForIdle())
        return ERROR_RESYNC;

    lsb = data&0xFF;
    msb = (data>>8) & 0xFF;
    
    uartPutData((uint8_t *)GENIE_WRITE_OBJ,16) ;
    checksum  = GENIE_WRITE_OBJ ;
    uartPutData((uint8_t *)&object,16) ;
    checksum ^= object ;
    uartPutData((uint8_t *)&index,16) ;
    checksum ^= index ;
    uartPutData((uint8_t *)&msb,16) ;
    checksum ^= msb;
    uartPutData((uint8_t *)&lsb,16) ;
    checksum ^= lsb;
    uartPutData(&checksum,16) ;
    
     return WaitForAnswer();
}

uint8_t genieWriteStr (uint16_t index, char *string)
{
        char *p ;
        unsigned int checksum ;
        int len = strlen (string) ;


        if (len > 255)
        return -1 ;


        //wait for interface to be ready before sending stuff
        if(WaitForIdle())
            return ERROR_RESYNC;


        uartPutData((uint8_t *)GENIE_WRITE_STR,16) ;    checksum  = GENIE_WRITE_STR ;
        uartPutData((uint8_t *)&index,16) ;              checksum ^= index ;
        uartPutChar((unsigned char)len) ; checksum ^= len ;
        for (p = string ; *p ; ++p)        {
        	uartPutData ((uint8_t *) p,8) ;
                checksum ^= *p ;
        }
        uartPutData((uint8_t *)&checksum,16) ;

        return  WaitForAnswer();
}
uint8_t  genieReadObj (uint16_t object, uint16_t index)
{
    //wait for interface to be ready before sending stuff
    if(WaitForIdle())
        return ERROR_RESYNC;
    unsigned int checksum ;
    
    uartPutData((uint8_t *)GENIE_READ_OBJ,16) ;     checksum  = GENIE_READ_OBJ ;
    uartPutData((uint8_t *)&object,16) ;             checksum ^= object ;
    uartPutData((uint8_t *)&index,16) ;              checksum ^= index ;
    uartPutData((uint8_t *)&checksum,16) ;
        
    //Here we don't wait for a typical answer
    //The screen will respond with an NACK if the command was not understood, otherwise it will send a report object frame    
    return  0;//WaitForReadAnswer();
}

void writec(char data)
{
	uartPutChar(data);
}

bool WaitForIdle()
{
    /*
        If the communication is still in progress
        lets wait for it to finish or check/wait for timeout
    */
    if(state != CommIdle)
    {
    	uint64_t ticks = RTCDRV_GetWallClockTicks64();
    	uint32_t milliseconds = RTCDRV_TicksToMsec(ticks);

    	long timeout = milliseconds + TIMEOUT_PERIOD;
        long timerReading = 0;
        if(milliseconds >= RxMaxTimeout)
        {
            Reset();
            RxStateTimeoutErrors++;
        }
        while(timerReading < timeout && state != CommIdle)
        {
            timerReading = milliseconds;
        }
        LastResponse = 0;
        return (timerReading >= timeout);
    }
    
    return false;    
}

/*
    Not used for now, since the screen will answer with either a report event or NAK
    Since those two answers are so different, there is no efficient means to receive
    both in interrupt. No wait to NAK will be done. DONT USE THIS FUNCTION ... YET
*/

int8_t WaitForReadAnswer()
{
    RTCDRV_StopTimer(_waitTimer);

    RTCDRV_StartTimer(_waitTimer, rtcdrvTimerTypeOneshot, 0, rtcCallback, NULL);

    uint64_t ticks = RTCDRV_GetWallClockTicks64();
    uint32_t milliseconds = RTCDRV_TicksToMsec(ticks);

    long timeout = milliseconds + TIMEOUT_PERIOD;
    long timerReading = 0;
    while(state == CommIdle && LastResponse != ERROR_NAK && timerReading < timeout)
    {
        timerReading = milliseconds;
    }

    if(LastResponse == ERROR_NAK)//check if the screen returned a NACK
    {
        LastResponse = NO_RESPONSE;
        return ERROR_NAK;
    }
    else if(milliseconds >= timeout) //check if we timed out while waiting for response
    {   
    
        LastResponse = NO_RESPONSE;
        return ERROR_TIMEOUT;
    }
    //if we get here it means we didnt timeout and the screen did accept the command
    LastResponse = NO_RESPONSE;
    return ERROR_NONE;
}



int8_t WaitForAnswer()
{
	uint64_t ticks = RTCDRV_GetWallClockTicks64();
	uint32_t milliseconds = RTCDRV_TicksToMsec(ticks);

    long timeout = milliseconds + TIMEOUT_PERIOD;

    long timerReading = 0;
    while(LastResponse != GENIE_ACK && LastResponse != ERROR_NAK && timerReading < timeout)
    {
        timerReading = milliseconds;
    }
           
    if(LastResponse == ERROR_NAK)
    {
        LastResponse = NO_RESPONSE;
        return ERROR_NAK;
    }
    else if(milliseconds >= timeout)
    {   
        printf("Current timer:%" PRIu32 "; timeout:%ld\n\r",milliseconds,timeout);
        LastResponse = NO_RESPONSE;
        return ERROR_TIMEOUT;
    }
    LastResponse = NO_RESPONSE;
    return ERROR_NONE;
}


void Reset(void)
{
    LastResponse = NO_RESPONSE;
    state = CommIdle;

/*    RTCDRV_StopTimer(_receptionTimer); */

    RTCDRV_StopTimer(_waitTimer);
    RTCDRV_StartTimer(_waitTimer, rtcdrvTimerTypeOneshot, 0, rtcCallback, NULL );

/*    RTCDRV_StartTimer(_receptionTimer, rtcdrvTimerTypeOneshot, 0, rtcCallback, NULL ); */

    GPIO_PinOutSet(gpioPortB, 11);

/*    do
    {
    	uartGetChar( );
    }
    while(uartGetChar( ));
*/
}


////////////////////// _genieFlushEventQueue ////////////////////
//
// Reset all the event queue variables and start from scratch.
//
void _genieFlushEventQueue(void) {
    _genieEventQueue.rd_index = 0;
    _genieEventQueue.wr_index = 0;
    _genieEventQueue.n_events = 0;
}

////////////////////// _genieEnqueueEvent ///////////////////
//
// Copy the bytes from a buffer supplied by the caller 
// to the input queue 
//
// Parms:        uint8_t * data, a pointer to the user's data
//
// Returns:        TRUE if there was an empty location in the queue
//                                to copy the data into
//                        FALSE if not
// Sets:        ERROR_REPLY_OVR if there was no room in the queue
//
bool _genieEnqueueEvent (uint8_t * data) {
    if (_genieEventQueue.n_events < MAX_GENIE_EVENTS-2) {
        memcpy (&_genieEventQueue.frames[_genieEventQueue.wr_index], data, GENIE_FRAME_SIZE);
        _genieEventQueue.wr_index++;
        _genieEventQueue.wr_index &= MAX_GENIE_EVENTS -1;
        _genieEventQueue.n_events++;
        return TRUE;
    } else {
        return FALSE;
    }
}
////////////////////// genieDequeueEvent ///////////////////
//
// Copy the bytes from a queued input event to a buffer supplied 
// by the caller.
//
// Parms:        genieFrame * buff, a pointer to the user's buffer
//
// Returns:        TRUE if there was an event to copy
//                        FALSE if not
//
bool genieDequeueEvent(union genieFrame * buff) {
    if (_genieEventQueue.n_events > 0) {
            memcpy (buff, &_genieEventQueue.frames[_genieEventQueue.rd_index], 
                            GENIE_FRAME_SIZE);
            _genieEventQueue.rd_index++;
            _genieEventQueue.rd_index &= MAX_GENIE_EVENTS -1;
            _genieEventQueue.n_events--;
            return TRUE;
    } 
    return FALSE;
}
void genieAttachEventHandler(genieUserEventHandlerPtr handler)
{
    _genieUserHandler = handler;
}
bool PendingFrames(void)
{
    return (_genieEventQueue.n_events>0);
}


/******************************************************************************
 * @brief  uartGetChar function
 *
 *  Note that if there are no pending characters in the receive buffer, this
 *  function will hang until a character is received.
 *
 *****************************************************************************/
uint8_t uartGetChar( )
{
  uint8_t ch;

  /* Check if there is a byte that is ready to be fetched. If no byte is ready, wait for incoming data */
  if (rxBuf.pendingBytes < 1)
  {
    while (rxBuf.pendingBytes < 1) ;
  }

  /* Copy data from buffer */
  ch        = rxBuf.data[rxBuf.rdI];
  rxBuf.rdI = (rxBuf.rdI + 1) % BUFFERSIZE;

  /* Decrement pending byte counter */
  rxBuf.pendingBytes--;

  return ch;
}




/******************************************************************************
 * @brief  uartPutChar function
 *
 *****************************************************************************/
void uartPutChar(uint8_t ch)
{
  /* Check if Tx queue has room for new data */
  if ((txBuf.pendingBytes + 1) > BUFFERSIZE)
  {
    /* Wait until there is room in queue */
    while ((txBuf.pendingBytes + 1) > BUFFERSIZE) ;
  }

  /* Copy ch into txBuffer */
  txBuf.data[txBuf.wrI] = ch;
  txBuf.wrI             = (txBuf.wrI + 1) % BUFFERSIZE;

  /* Increment pending byte counter */
  txBuf.pendingBytes++;

  /* Enable interrupt on USART TX Buffer*/
  USART_IntEnable(uart, UART_IEN_TXBL);
}




/******************************************************************************
 * @brief  uartPutData function
 *
 *****************************************************************************/
void uartPutData(uint8_t * dataPtr, uint32_t dataLen)
{
  uint32_t i = 0;

  /* Check if buffer is large enough for data */
  if (dataLen > BUFFERSIZE)
  {
    /* Buffer can never fit the requested amount of data */
    return;
  }

  /* Check if buffer has room for new data */
  if ((txBuf.pendingBytes + dataLen) > BUFFERSIZE)
  {
    /* Wait until room */
    while ((txBuf.pendingBytes + dataLen) > BUFFERSIZE) ;
  }

  /* Fill dataPtr[0:dataLen-1] into txBuffer */
  while (i < dataLen)
  {
    txBuf.data[txBuf.wrI] = *(dataPtr + i);
    txBuf.wrI             = (txBuf.wrI + 1) % BUFFERSIZE;
    i++;
  }

  /* Increment pending byte counter */
  txBuf.pendingBytes += dataLen;

  /* Enable interrupt on USART TX Buffer*/
  USART_IntEnable(uart, UART_IEN_TXBL);
}

/******************************************************************************
 * @brief  uartGetData function
 *
 *****************************************************************************/
uint32_t uartGetData(uint8_t * dataPtr, uint32_t dataLen)
{
  uint32_t i = 0;

  /* Wait until the requested number of bytes are available */
  if (rxBuf.pendingBytes < dataLen)
  {
    while (rxBuf.pendingBytes < dataLen) ;
  }

  if (dataLen == 0)
  {
    dataLen = rxBuf.pendingBytes;
  }

  /* Copy data from Rx buffer to dataPtr */
  while (i < dataLen)
  {
    *(dataPtr + i) = rxBuf.data[rxBuf.rdI];
    rxBuf.rdI      = (rxBuf.rdI + 1) % BUFFERSIZE;
    i++;
  }

  /* Decrement pending byte counter */
  rxBuf.pendingBytes -= dataLen;

  return i;
}




/**************************************************************************//**
 * @brief UART1 RX IRQ Handler
 *
 * Set up the interrupt prior to use
 *
 * Note that this function handles overflows in a very simple way.
 *
 *****************************************************************************/
void UART1_RX_IRQHandler(void)
{
  /* Check for RX data valid interrupt */
  if (uart->IF & UART_IF_RXDATAV)
  {
    /* Copy data into RX Buffer */
    uint8_t rxData = USART_Rx(uart);
    rxBuf.data[rxBuf.wrI] = rxData;
    rxBuf.wrI             = (rxBuf.wrI + 1) % BUFFERSIZE;
    rxBuf.pendingBytes++;

    uint8_t c;

    uint64_t ticks = RTCDRV_GetWallClockTicks64();
    uint32_t milliseconds = RTCDRV_TicksToMsec(ticks);

   	do
	{
		c = uartGetChar();

		/* check if we are in reception mode and if we timeout */
		if(milliseconds >= RxMaxTimeout && state == CommInProgress)
		{
			/* if we get here, it means something bad has happened and
			we need to reset communication status */
			Reset();
		}

		ManageReceiveData(c);
	}
	while(uartGetChar( ));



    /* Flag Rx overflow
    if (rxBuf.pendingBytes > BUFFERSIZE)
    {
      rxBuf.overflow = true;
    }
    */
  }
}

/**************************************************************************//**
 * @brief UART1 TX IRQ Handler
 *
 * Set up the interrupt prior to use
 *
 *****************************************************************************/
void UART1_TX_IRQHandler(void)
{
  /* Check TX buffer level status */
  if (uart->IF & UART_IF_TXBL)
  {
    if (txBuf.pendingBytes > 0)
    {
      /* Transmit pending character */
      USART_Tx(uart, txBuf.data[txBuf.rdI]);
      txBuf.rdI = (txBuf.rdI + 1) % BUFFERSIZE;
      txBuf.pendingBytes--;
    }

    /* Disable Tx interrupt if no more bytes in queue */
    if (txBuf.pendingBytes == 0)
    {
      USART_IntDisable(uart, UART_IEN_TXBL);
    }
  }
}

